#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Itay Marom
Cisco Systems, Inc.

Copyright(c) 2015-2015 Cisco Systems, Inc.
Licensed under the Apache License, Version 2.0(the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import argparse
import json
import os
import subprocess
import sys
import tempfile

# simulator can be run as a standalone
from trex_stl_lib.trex_stl_exceptions import STLArgumentError, STLError
from trex_stl_lib.trex_stl_packet_builder_scapy import RawPcapReader
from trex_stl_lib.trex_stl_streams import STLProfile, STLStream
from trex_stl_lib.utils import parsing_opts, pcap
from trex_stl_lib.utils.text_opts import format_text


class BpSimException(Exception):
    pass


# stateless simulation
class STLSim(object):
    MASK_ALL = ((1 << 64) - 1)

    def __init__(self, bp_sim_path, handler=0, port_id=0, api_h="dummy"):

        self.bp_sim_path = os.path.abspath(bp_sim_path)
        if not os.path.exists(self.bp_sim_path):
            raise STLError('BP sim path %s does not exist' % self.bp_sim_path)

        # dummies
        self.handler = handler
        self.api_h = api_h
        self.port_id = port_id

    def generate_start_cmd(self, mult="1", force=True, duration=-1):
        return {"id": 1,
                "jsonrpc": "2.0",
                "method": "start_traffic",
                "params": {"handler": self.handler,
                           "api_h": self.api_h,
                           "force":  force,
                           "port_id": self.port_id,
                           "mul": parsing_opts.decode_multiplier(mult),
                           "duration": duration,
                           "core_mask": self.MASK_ALL}
                }

    # run command
    # input_list - a list of streams or YAML files
    # outfile - pcap file to save output, if None its a dry run
    # dp_core_count - how many DP cores to use
    # dp_core_index - simulate only specific dp core without merging
    # is_debug - debug or release image
    # pkt_limit - how many packets to simulate
    # mult - multiplier
    # mode - can be 'valgrind, 'gdb', 'json' or 'none'

    def run(self,
            input_list,
            outfile=None,
            dp_core_count=1,
            dp_core_index=None,
            is_debug=True,
            pkt_limit=5000,
            mult="1",
            duration=-1,
            mode='none',
            silent=False,
            tunables=None):

        if mode not in ['none', 'gdb', 'valgrind', 'json', 'yaml', 'pkt', 'native']:
            raise STLArgumentError('mode', mode)

        # listify
        input_list = input_list if isinstance(
            input_list, list) else [input_list]

        # check streams arguments
        if not all([isinstance(i, (STLStream, str)) for i in input_list]):
            raise STLArgumentError('input_list', input_list)

        # split to two type
        input_files = [x for x in input_list if isinstance(x, str)]
        stream_list = [x for x in input_list if isinstance(x, STLStream)]

        # handle YAMLs
        if tunables is None:
            tunables = {}

        for input_file in input_files:
            try:
                if 'direction' not in tunables:
                    tunables['direction'] = self.port_id % 2

                profile = STLProfile.load(input_file, **tunables)

            except STLError as e:
                s = format_text(
                    "\nError while loading profile '{0}'\n".format(input_file), 'bold')
                s += "\n" + e.brief()
                raise STLError(s)

            stream_list += profile.get_streams()

        # load streams
        cmds_json = []

        id_counter = 1

        lookup = {}

        # allocate IDs
        for stream in stream_list:
            if stream.get_id() is not None:
                stream_id = stream.get_id()
            else:
                stream_id = id_counter
                id_counter += 1

            name = stream.get_name() if stream.get_name() is not None else id(stream)
            if name in lookup:
                raise STLError(
                    "multiple streams with name: '{0}'".format(name))
            lookup[name] = stream_id

        # resolve names
        for stream in stream_list:

            name = stream.get_name() if stream.get_name() is not None else id(stream)
            stream_id = lookup[name]

            next_id = -1
            next = stream.get_next()
            if next:
                if next not in lookup:
                    raise STLError(
                        "stream dependency error - unable to find '{0}'".format(next))
                next_id = lookup[next]

            stream_json = stream.to_json()
            stream_json['next_stream_id'] = next_id

            cmd = {"id": 1,
                   "jsonrpc": "2.0",
                   "method": "add_stream",
                   "params": {"handler": self.handler,
                              "api_h": self.api_h,
                              "port_id": self.port_id,
                              "stream_id": stream_id,
                              "stream": stream_json}
                   }

            cmds_json.append(cmd)

        # generate start command
        cmds_json.append(self.generate_start_cmd(mult=mult,
                                                 force=True,
                                                 duration=duration))

        if mode == 'json':
            print((json.dumps(cmds_json, indent=4,
                              separators=(',', ': '), sort_keys=True)))
            return
        elif mode == 'yaml':
            print((STLProfile(stream_list).dump_to_yaml()))
            return
        elif mode == 'pkt':
            print((STLProfile(stream_list).dump_as_pkt()))
            return
        elif mode == 'native':
            print((STLProfile(stream_list).dump_to_code()))
            return

        # start simulation
        self.outfile = outfile
        self.dp_core_count = dp_core_count
        self.dp_core_index = dp_core_index
        self.is_debug = is_debug
        self.pkt_limit = pkt_limit
        self.mult = mult
        self.duration = duration,
        self.mode = mode
        self.silent = silent

        self.__run(cmds_json)

    # internal run

    def __run(self, cmds_json):

        # write to temp file
        f = tempfile.NamedTemporaryFile(delete=False)

        msg = json.dumps(cmds_json).encode()

        f.write(msg)
        f.close()

        # launch bp-sim
        try:
            self.execute_bp_sim(f.name)
        finally:
            os.unlink(f.name)

    def execute_bp_sim(self, json_filename):
        if self.is_debug:
            exe = os.path.join(self.bp_sim_path, 'bp-sim-64-debug')
        else:
            exe = os.path.join(self.bp_sim_path, 'bp-sim-64')

        if not os.path.exists(exe):
            raise STLError(
                "'{0}' does not exists, please build it before calling the simulation".format(exe))

        cmd = [exe,
               '--pcap',
               '--sl',
               '--cores',
               str(self.dp_core_count),
               '--limit',
               str(self.pkt_limit),
               '-f',
               json_filename]

        # out or dry
        if not self.outfile:
            cmd += ['--dry']
            cmd += ['-o', '/dev/null']
        else:
            cmd += ['-o', self.outfile]

        if self.dp_core_index is not None:
            cmd += ['--core_index', str(self.dp_core_index)]

        if self.mode == 'valgrind':
            cmd = ['valgrind', '--leak-check=full', '--error-exitcode=1'] + cmd

        elif self.mode == 'gdb':
            cmd = ['/usr/bin/gdb', '--args'] + cmd

        print(("executing command: '{0}'".format(" ".join(cmd))))

        if self.silent:
            FNULL = open(os.devnull, 'wb')
            rc = subprocess.call(cmd, stdout=FNULL)
        else:
            rc = subprocess.call(cmd)

        if rc != 0:
            raise STLError(
                'simulation has failed with error code {0}'.format(rc))

        self.merge_results()

    def merge_results(self):
        if not self.outfile:
            return

        if self.dp_core_count == 1:
            return

        if self.dp_core_index is not None:
            return

        if not self.silent:
            print("Mering cores output to a single pcap file...\n")
        inputs = ["{0}-{1}".format(self.outfile, index)
                  for index in range(0, self.dp_core_count)]
        pcap.merge_cap_files(inputs, self.outfile, delete_src=True)


def is_valid_file(filename):
    if not os.path.isfile(filename):
        raise argparse.ArgumentTypeError(
            "The file '%s' does not exist" % filename)

    return filename


def unsigned_int(x):
    x = int(x)
    if x < 0:
        raise argparse.ArgumentTypeError("argument must be >= 0")

    return x


def setParserOptions():
    parser = argparse.ArgumentParser(prog="stl_sim.py")

    parser.add_argument("-f",
                        dest="input_file",
                        help="input file in YAML or Python format",
                        type=is_valid_file,
                        required=True)

    parser.add_argument("-o",
                        dest="output_file",
                        default=None,
                        help="output file in ERF format")

    parser.add_argument("-c", "--cores",
                        help="DP core count [default is 1]",
                        dest="dp_core_count",
                        default=1,
                        type=int,
                        choices=list(range(1, 9)))

    parser.add_argument("-n", "--core_index",
                        help="Record only a specific core",
                        dest="dp_core_index",
                        default=None,
                        type=int)

    parser.add_argument("-i", "--port",
                        help="Simulate a specific port ID [default is 0]",
                        dest="port_id",
                        default=0,
                        type=int)

    parser.add_argument("-r", "--release",
                        help="runs on release image instead of debug [default is False]",
                        action="store_true",
                        default=False)

    parser.add_argument("-s", "--silent",
                        help="runs on silent mode(no stdout) [default is False]",
                        action="store_true",
                        default=False)

    parser.add_argument("-l", "--limit",
                        help="limit test total packet count [default is 5000]",
                        default=5000,
                        type=unsigned_int)

    parser.add_argument('-m', '--multiplier',
                        help=parsing_opts.match_multiplier_help,
                        dest='mult',
                        default="1",
                        type=parsing_opts.match_multiplier_strict)

    parser.add_argument('-d', '--duration',
                        help="run duration",
                        dest='duration',
                        default=-1,
                        type=float)

    parser.add_argument('-t',
                        help='sets tunable for a profile',
                        dest='tunables',
                        default=None,
                        type=parsing_opts.decode_tunables)

    parser.add_argument('-p', '--path',
                        help="BP sim path",
                        dest='bp_sim_path',
                        default=None,
                        type=str)

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-x', '--valgrind',
                       help='run under valgrind [default is False]',
                       action='store_true',
                       default=False)

    group.add_argument('-g', '--gdb',
                       help='run under GDB [default is False]',
                       action='store_true',
                       default=False)

    group.add_argument('--json',
                       help='generate JSON output only to stdout [default is False]',
                       action='store_true',
                       default=False)

    group.add_argument('--pkt',
                       help='Parse the packet and show it as hex',
                       action='store_true',
                       default=False)

    group.add_argument('--yaml',
                       help='generate YAML from input file [default is False]',
                       action='store_true',
                       default=False)

    group.add_argument('--native',
                       help=('generate Python code with stateless profile '
                             + 'from input file [default is False]'),
                       action='store_true',
                       default=False)

    group.add_argument('--test_multi_core',
                       help='runs the profile with c=1-8',
                       action='store_true',
                       default=False)

    return parser


def validate_args(parser, options):

    if options.dp_core_index:
        if options.dp_core_index not in range(0, options.dp_core_count):
            parser.error("DP core index valid range is 0 to {0}".format(
                options.dp_core_count - 1))

    # zero is ok - no limit, but other values must be at least as the number of cores
    if(options.limit != 0) and options.limit < options.dp_core_count:
        parser.error("limit cannot be lower than number of DP cores")


# a more flexible check
def compare_caps(cap1, cap2, max_diff_sec=(5 * 1e-6)):
    pkts1 = list(RawPcapReader(cap1))
    pkts2 = list(RawPcapReader(cap2))

    if len(pkts1) != len(pkts2):
        print(('{0} contains {1} packets vs. {2} contains {3} packets'.format(
            cap1, len(pkts1), cap2, len(pkts2))))
        return False

    # to be less strict we define equality if all packets from cap1 exists and in cap2
    # and vice versa
    # 'exists' means the same packet with abs(TS1-TS2) < 5nsec
    # its O(n^2) but who cares, right ?
    for i, pkt1 in enumerate(pkts1):
        ts1 = float(pkt1[1][0]) + (float(pkt1[1][1]) / 1e6)
        found = None
        for j, pkt2 in enumerate(pkts2):
            ts2 = float(pkt2[1][0]) + (float(pkt2[1][1]) / 1e6)

            if abs(ts1-ts2) > max_diff_sec:
                break

            if pkt1[0] == pkt2[0]:
                found = j
                break

        if found is None:
            print((format_text(
                "cannot find packet #{0} from {1} in {2}\n".format(i, cap1, cap2), 'bold')))
            return False
        else:
            del pkts2[found]

    return True


def hexdiff(d1, d2):
    rc = []

    if len(d1) != len(d2):
        return rc

    for i in range(len(d1)):
        if d1[i] != d2[i]:
            rc.append(i)
    return rc


def prettyhex(h, diff_list):
    if type(h[0]) == str:
        h = [ord(x) for x in h]

    for i in range(len(h)):

        if i in diff_list:
            sys.stdout.write("->'0x%02x'<-" % h[i])
        else:
            sys.stdout.write("  '0x%02x'  " % h[i])
        if((i % 9) == 8):
            print("")

    print("")

# a more strict comparsion 1 <--> 1


def compare_caps_strict(cap1, cap2, max_diff_sec=(5 * 1e-6)):
    pkts1 = list(RawPcapReader(cap1))
    pkts2 = list(RawPcapReader(cap2))

    if len(pkts1) != len(pkts2):
        print(('{0} contains {1} packets vs. {1} contains {2} packets'.format(
            cap1, len(pkts1), cap2, len(pkts2))))
        return False

    # a strict check
    for pkt1, pkt2, i in zip(pkts1, pkts2, list(range(1, len(pkts1)))):
        ts1 = float(pkt1[1][0]) + (float(pkt1[1][1]) / 1e6)
        ts2 = float(pkt2[1][0]) + (float(pkt2[1][1]) / 1e6)

        if abs(ts1-ts2) > 0.000005:  # 5 nsec
            print((format_text(
                "TS error: cap files '{0}', '{1}' differ in cap #{2} - '{3}' vs. '{4}'\n"
                .format(cap1, cap2, i, ts1, ts2), 'bold')))
            return False

        if pkt1[0] != pkt2[0]:
            print((format_text("RAW error: cap files '{0}', '{1}' differ in cap #{2}\n".format(
                cap1, cap2, i), 'bold')))

            diff_list = hexdiff(pkt1[0], pkt2[0])

            print(("{0} - packet #{1}:\n".format(cap1, i)))
            prettyhex(pkt1[0], diff_list)

            print(("\n{0} - packet #{1}:\n".format(cap2, i)))
            prettyhex(pkt2[0], diff_list)

            print("")
            return False

    return True


def test_multi_core(r, options):

    for core_count in range(1, 9):
        r.run(input_list=options.input_file,
              outfile='generated/{0}.cap'.format(core_count),
              dp_core_count=core_count,
              is_debug=(not options.release),
              pkt_limit=options.limit,
              mult=options.mult,
              duration=options.duration,
              mode='none',
              silent=True,
              tunables=options.tunables)

    print("")

    for core_count in range(1, 9):
        print((format_text("comparing {0} cores to 1 core:\n".format(
            core_count), 'underline')))
        rc = compare_caps_strict(
            'generated/1.cap', 'generated/{0}.cap'.format(core_count))
        if rc:
            print("[Passed]\n")

    return


def main(args=None):
    parser = setParserOptions()
    options = parser.parse_args(args=args)

    validate_args(parser, options)

    if options.valgrind:
        mode = 'valgrind'
    elif options.gdb:
        mode = 'gdb'
    elif options.json:
        mode = 'json'
    elif options.yaml:
        mode = 'yaml'
    elif options.native:
        mode = 'native'
    elif options.pkt:
        mode = 'pkt'
    elif options.test_multi_core:
        mode = 'test_multi_core'
    else:
        mode = 'none'

    try:
        r = STLSim(bp_sim_path=options.bp_sim_path, port_id=options.port_id)

        if mode == 'test_multi_core':
            test_multi_core(r, options)
        else:
            r.run(input_list=options.input_file,
                  outfile=options.output_file,
                  dp_core_count=options.dp_core_count,
                  dp_core_index=options.dp_core_index,
                  is_debug=(not options.release),
                  pkt_limit=options.limit,
                  mult=options.mult,
                  duration=options.duration,
                  mode=mode,
                  silent=options.silent,
                  tunables=options.tunables)

    except KeyboardInterrupt:
        print("\n\n*** Caught Ctrl + C... Exiting...\n\n")
        return(-1)

    except STLError as e:
        print(e)
        return(-1)

    return(0)


if __name__ == '__main__':
    main()



#!/usr/bin/python3

import json
import re
from threading import Lock

import zmq
from trex_stl_lib.trex_stl_types import RC, RC_ERR, RC_OK
from trex_stl_lib.utils.common import random_id_gen
from trex_stl_lib.utils.zipmsg import ZippedMsg


class bcolors:
    BLUE = '\033[94m'
    GREEN = '\033[32m'
    YELLOW = '\033[93m'
    RED = '\033[31m'
    MAGENTA = '\033[35m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

# sub class to describe a batch


class BatchMessage(object):
    def __init__(self, rpc_client):
        self.rpc_client = rpc_client
        self.batch_list = []

    def add(self, method_name, params=None, api_class='core'):

        id, msg = self.rpc_client.create_jsonrpc_v2(
            method_name, params, api_class, encode=False)
        self.batch_list.append(msg)

    def invoke(self, block=False, chunk_size=500000, retry=0):
        if not self.rpc_client.connected:
            return RC_ERR("Not connected to server")

        if chunk_size:
            response_batch = RC()
            size = 0
            new_batch = []
            for msg in self.batch_list:
                size += len(json.dumps(msg))
                new_batch.append(msg)

                if size > chunk_size:
                    batch_json = json.dumps(new_batch)
                    response_batch.add(self.rpc_client.send_msg(batch_json))
                    size = 0
                    new_batch = []
            if new_batch:
                batch_json = json.dumps(new_batch)
                response_batch.add(self.rpc_client.send_msg(batch_json))
            return response_batch
        else:
            batch_json = json.dumps(self.batch_list)
            return self.rpc_client.send_msg(batch_json, retry=retry)


# JSON RPC v2.0 client
class JsonRpcClient(object):

    def __init__(self, default_server, default_port, client):
        self.get_api_h = client._get_api_h
        self.logger = client.logger
        self.connected = False

        # default values
        self.port = default_port
        self.server = default_server

        self.id_gen = random_id_gen()
        self.zipper = ZippedMsg()

        self.lock = Lock()

    def get_connection_details(self):
        rc = {}
        rc['server'] = self.server
        rc['port'] = self.port

        return rc

    # pretty print for JSON
    def pretty_json(self, json_str, use_colors=True):
        pretty_str = json.dumps(json.loads(
            json_str), indent=4, separators=(',', ': '), sort_keys=True)

        if not use_colors:
            return pretty_str

        try:
            # int numbers
            pretty_str = re.sub(r'([ ]*:[ ]+)(\-?[1-9][0-9]*[^.])',
                                r'\1{0}\2{1}'.format(bcolors.BLUE, bcolors.ENDC), pretty_str)
            # float
            pretty_str = re.sub(r'([ ]*:[ ]+)(\-?[1-9][0-9]*\.[0-9]+)',
                                r'\1{0}\2{1}'.format(bcolors.MAGENTA, bcolors.ENDC), pretty_str)
            # strings

            pretty_str = re.sub(
                r'([ ]*:[ ]+)("[^"]*")', r'\1{0}\2{1}'.format(bcolors.RED,
                                                              bcolors.ENDC), pretty_str)
            pretty_str = re.sub(
                r"('[^']*')", r'{0}\1{1}'.format(bcolors.MAGENTA, bcolors.RED), pretty_str)
        except Exception:
            pass

        return pretty_str

    def verbose_msg(self, msg):
        self.logger.log("\n\n[verbose] " + msg, level=self.logger.VERBOSE_HIGH)

    # batch messages

    def create_batch(self):
        return BatchMessage(self)

    def create_jsonrpc_v2(self, method_name, params=None, api_class='core', encode=True):
        msg = {}
        msg["jsonrpc"] = "2.0"
        msg["method"] = method_name
        msg["id"] = next(self.id_gen)
        msg["params"] = params if params is not None else {}

        # if this RPC has an API class - add it's handler
        if api_class:
            msg["params"]["api_h"] = self.get_api_h()[api_class]

        # pdb.set_trace()
        if encode:
            return id, json.dumps(msg)
        else:
            return id, msg

    def invoke_rpc_method(self, method_name, params=None, api_class='core', retry=0):
        if not self.connected:
            return RC_ERR("Not connected to server")

        id, msg = self.create_jsonrpc_v2(method_name, params, api_class)

        return self.send_msg(msg, retry=retry)

    def send_msg(self, msg, retry=0):
        # REQ/RESP pattern in ZMQ requires no interrupts during the send
        with self.lock:
            return self.__send_msg(msg, retry)

    def __send_msg(self, msg, retry=0):
        # print before
        if self.logger.check_verbose(self.logger.VERBOSE_HIGH):
            self.verbose_msg("Sending Request To Server:\n\n" +
                             self.pretty_json(msg) + "\n")

        # encode string to buffer
        buffer = msg.encode("utf-8", "ignore")

        if self.zipper.check_threshold(buffer):
            response = self.send_raw_msg(
                self.zipper.compress(buffer), retry=retry)
        else:
            response = self.send_raw_msg(buffer, retry=retry)

        if not response:
            return response
        elif self.zipper.is_compressed(response):
            response = self.zipper.decompress(response)

        # return to string
        response = response.decode()

        # print after
        if self.logger.check_verbose(self.logger.VERBOSE_HIGH):
            self.verbose_msg("Server Response:\n\n" +
                             self.pretty_json(response) + "\n")

        # process response(batch and regular)
        try:
            response_json = json.loads(response)
        except(TypeError, ValueError):
            return RC_ERR("*** [RPC] - Failed to decode response from server")

        if isinstance(response_json, list):
            return self.process_batch_response(response_json)
        else:
            return self.process_single_response(response_json)

    # low level send of string message

    def send_raw_msg(self, msg, retry=0):
        try:
            return self._send_raw_msg_safe(msg, retry)
        except KeyboardInterrupt as e:
            # must restore the socket to a sane state
            self.reconnect()
            raise e

    def _send_raw_msg_safe(self, msg, retry):

        retry_left = retry
        while True:
            try:
                # pdb.set_trace()
                if isinstance(msg, str):
                    msg = msg  # .encode("utf-8", "ignore")

                if isinstance(msg, bytes):
                    self.socket.send(msg)
                else:
                    return RC_ERR("*** [RPC] - failed to understand message to server")
                break
            except zmq.Again:
                retry_left -= 1
                if retry_left < 0:
                    self.disconnect()
                    return RC_ERR("*** [RPC] - Failed to send message to server")

        retry_left = retry
        while True:
            try:
                response = self.socket.recv()
                break
            except zmq.Again:
                retry_left -= 1
                if retry_left < 0:
                    self.disconnect()
                    return RC_ERR("*** [RPC] - Failed to get server response from {0}"
                                  .format(self.transport))

        return response

    # processs a single response from server

    def process_single_response(self, response_json):

        if(response_json.get("jsonrpc") != "2.0"):
            return RC_ERR("Malformed Response({0})".format(str(response_json)))

        # error reported by server
        if("error" in response_json):
            if "specific_err" in response_json["error"]:
                return RC_ERR(response_json["error"]["specific_err"])
            else:
                return RC_ERR(response_json["error"]["message"])

        # if no error there should be a result
        if("result" not in response_json):
            return RC_ERR("Malformed Response({0})".format(str(response_json)))

        return RC_OK(response_json["result"])

    # process a batch response

    def process_batch_response(self, response_json):
        rc_batch = RC()

        for single_response in response_json:
            rc = self.process_single_response(single_response)
            rc_batch.add(rc)

        return rc_batch

    def disconnect(self):
        if self.connected:
            self.socket.close(linger=0)
            self.context.destroy(linger=0)
            self.connected = False
            return RC_OK()
        else:
            return RC_ERR("Not connected to server")

    def connect(self, server=None, port=None):
        if self.connected:
            self.disconnect()

        self.context = zmq.Context()

        self.server = (server if server else self.server)
        self.port = (port if port else self.port)

        #  Socket to talk to server
        self.transport = "tcp://{0}:{1}".format(self.server, self.port)

        self.socket = self.context.socket(zmq.REQ)
        try:
            self.socket.connect(self.transport)
        except zmq.error.ZMQError as e:
            return RC_ERR("ZMQ Error: Bad server or port name: " + str(e))

        self.socket.setsockopt(zmq.SNDTIMEO, 10000)
        self.socket.setsockopt(zmq.RCVTIMEO, 10000)

        self.connected = True

        return RC_OK()

    def reconnect(self):
        # connect using current values
        return self.connect()

    def is_connected(self):
        return self.connected

    def __del__(self):
        self.logger.log(msg="Shutting down RPC client\n")
        if hasattr(self, "context"):
            self.context.destroy(linger=0)



#!/usr/bin/python3

"""
Internal objects for service implementation

Description:
  Internal objects used by the library to implement
  service capabilities

  Objects from this file should not be
  directly created by the user

Author:
  Itay Marom

"""

from collections import deque, namedtuple

import simpy
from simpy.core import BoundClass

from trex_stl_lib.trex_stl_exceptions import STLError

from trex_stl_lib.trex_stl_psv import PSV_ACQUIRED, PSV_SERVICE, PSV_UP

from .trex_stl_service import STLService

##################
#
# STL service context
#
#
##################


class STLServiceCtx(object):
    '''
        service context provides the
        envoirment for running many services
        and their spawns in parallel
    '''

    def __init__(self, client, port):
        self.client = client
        self.port = port
        self.port_obj = client.ports[port]
        self._reset()

# ######## API functions              #########

    def run(self, services, pps=1000):
        '''
            Runs 'services' under service context

            'pps' - provides a rate for services to
                    generate traffic
        '''

        if not services:
            raise Exception('No services to run!')

        self.pps = pps
        self.ipg_sec = 1.0 / pps
        self.ipg_usec = int(self.ipg_sec * 1e6)

        # for TX batch size, we cannot accumulate more than 1 second
        self.tx_batch_size = min(100, self.pps)

        with self.client.logger.supress():
            self._run(services)

    def get_port_id(self):
        '''
            Returns the port ID attached to
            the context
        '''
        return self.port

    def get_src_ipv4(self):
        '''
            Returns the source IPv4 of
            the port under the context
            or None if the port is configured as L2
        '''
        layer_cfg = self.port_obj.get_layer_cfg()
        return layer_cfg['ipv4']['src'] if self.port_obj.is_l3_mode() else None

    def get_src_mac(self):
        '''
            returns the SRC mac of the port
            attached to the service
        '''

        layer_cfg = self.port_obj.get_layer_cfg()
        return layer_cfg['ether']['src']

# ######## internal functions              #########

    def _reset(self):
        self.filters = {}
        self.services = {}
        self.done_services = 0

    def _add(self, services):
        '''
            Add a service to the context
        '''
        if isinstance(services, STLService):
            self._add_single_service(services)

        elif isinstance(services, (list, tuple)) and all([isinstance(s, STLService)
                                                         for s in services]):
            for service in services:
                self._add_single_service(service)

        else:
            raise STLError(
                "'services' should be STLService subtype or list/tuple of it")

    def _run(self, services):

        # check port state
        self.client.psv.validate('SERVICE CTX', ports=self.port,
                                 states=(PSV_UP,
                                         PSV_ACQUIRED,
                                         PSV_SERVICE))

        # prepare
        self._reset()

        # add all services
        self._add(services)

        # create an environment
        self.env = simpy.rt.RealtimeEnvironment(factor=1, strict=False)
        self.tx_buffer = TXBuffer(
            self.env, self.client, self.port, self.tx_batch_size, self.ipg_usec)

        # start all services
        for service in self.services:
            pipe = self._pipe()
            self.services[service]['pipe'] = pipe
            p = self.env.process(service.run(pipe))
            self._on_process_create(p)

        # save promisicous state and move to enabled
        is_promiscuous = self.client.get_port_attr(port=self.port)[
            'prom'] == "on"
        if not is_promiscuous:
            self.client.set_port_attr(ports=self.port, promiscuous=True)

        try:
            # for each filter, start a capture
            for f in list(self.filters.values()):

                # capture limit is dictated by the PPS times a factor
                limit = max(self.pps * 5, 10000)
                f['capture_id'] = self.client.start_capture(
                    rx_ports=self.port, bpf_filter=f['inst'].get_bpf_filter(), limit=limit)['id']

            # add maintence processes
            tx_process = self.env.process(self._tx_pkts_process())
            self.env.process(self._rx_pkts_process())

            # start the RT simulation - exit when the TX process has ended
            # (RX has no sense if no proceesses exists)
            self.env.run(until=tx_process)

        finally:
            # stop all captures
            for f in list(self.filters.values()):
                if f['capture_id'] is not None:
                    self.client.stop_capture(f['capture_id'])

            if not is_promiscuous:
                self.client.set_port_attr(ports=self.port, promiscuous=False)
            self._reset()

    def _add_single_service(self, service):

        filter_type = service.get_filter_type()

        # if the service does not have a filter installed - create it
        if filter_type not in self.filters:
            self.filters[filter_type] = {
                'inst': filter_type(), 'capture_id': None}

        # add to the filter
        self.filters[filter_type]['inst'].add(service)

        # data per service
        self.services[service] = {'pipe': None}

    def _on_process_create(self, p):
        p.callbacks.append(self._on_process_exit)

    def _on_process_exit(self, event):
        self.done_services += 1

    def _pipe(self):
        return STLServicePipe(self.env, self.tx_buffer)

    def _fetch_rx_pkts_per_filter(self, f):
        pkts = []
        self.client.fetch_capture_packets(f['capture_id'], pkts)

        # for each packet - try to forward to each service until we hit
        for pkt in pkts:
            pkt_bin = pkt['binary']
            rx_ts = pkt['ts']

            # lookup all the services that this filter matches(usually 1)
            services = f['inst'].lookup(pkt_bin)

            for service in services:
                self.services[service]['pipe']._on_rx_pkt(pkt_bin, rx_ts)

        return len(pkts)

    def is_done(self):
        return self.done_services == len(self.services)

    # TX pkts process

    def _tx_pkts_process(self):

        sent = 0

        while not self.is_done():
            interval_sec = self.ipg_sec * sent if sent else 0.1
            yield self.env.timeout(interval_sec)

            sent = self.tx_buffer.send_all()

    # reads packets from the server

    def _rx_pkts_process(self):
        while not self.is_done():

            yield self.env.timeout(0.1)

            for f in list(self.filters.values()):
                self._fetch_rx_pkts_per_filter(f)


class TXBuffer(object):
    '''
        TX buffer
        handles buffering and sending packets
    '''

    # defines a basic buffer to hold packets to be sent
    Buffer = namedtuple('Buffer', ['pkts', 'event'])

    def __init__(self, env, client, port, threshold, ipg_usec):
        self.env = env
        self.client = client
        self.port = port
        self.threshold = threshold
        self.ipg_usec = ipg_usec

        # fast queue
        self.queue = deque()

    @property
    def head(self):
        '''
            returns the head of the queue - a buffer
        '''
        return self.queue[-1]

    def allocate_new_buffer(self):
        '''
            allocate a new buffer and append it
        '''

        buffer = TXBuffer.Buffer(pkts=[], event=self.env.event())
        self.queue.append(buffer)

    def roomleft(self):
        '''
            return true if there is still room in the current buffer
        '''
        return self.queue and(len(self.head.pkts) < self.threshold)

    def push(self, pkt):
        '''
            push a packet to the head buffer
            returns the event assosicated with it
        '''

        # if no room left, allocate a new queue
        if not self.roomleft():
            self.allocate_new_buffer()

        # should always havea room after new buffer
        assert(self.roomleft())

        self.head.pkts.append(pkt)

        return self.head.event

    def send_all(self):
        '''
            performs a single tick - one buffer will be sent(max of THRESHOLD)
        '''

        if self.queue:
            buffer = self.queue.popleft()
            rc = self.client.push_packets(
                ports=self.port, pkts=buffer.pkts, ipg_usec=self.ipg_usec, force=True)
            tx_ts = rc.data()['ts']

            self.pkts = []

            # mark as TX event happened
            buffer.event.succeed(value={'ts': tx_ts})

            return len(buffer.pkts)

        else:
            return 0


class PktRX(simpy.resources.store.StoreGet):
    '''
        An event waiting for RX packets

        'limit' - the limit for the get event
                  None means unlimited
    '''

    def __init__(self, store, timeout_sec=None, limit=None):
        self.limit = limit

        if timeout_sec is not None:
            self.timeout = store._env.timeout(timeout_sec)
            self.timeout.callbacks.append(self.on_get_timeout)

        super(PktRX, self).__init__(store)

    def on_get_timeout(self, event):
        '''
            Called when a timeout for RX packet has occured
            The event will be cancled(removed from queue)
            and a None value will be returend
        '''
        if not self.triggered:
            self.cancel()
            self.succeed([])


class Pkt(simpy.resources.store.Store):

    get = BoundClass(PktRX)

    def _do_get(self, event):
        if self.items:

            # if no limit - fetch all
            if event.limit is None:
                event.succeed(self.items)
                self.items = []

            else:
                # if limit was set - slice the list
                event.succeed(self.items[:event.limit])
                self.items = self.items[event.limit:]


##################
#
# STL service pipe
#
#
##################

class STLServicePipe(object):
    '''
        A pipe used to communicate between
        a service and the infrastructure
    '''

    def __init__(self, env, tx_buffer):
        self.env = env
        self.tx_buffer = tx_buffer
        self.pkt = Pkt(self.env)

    def async_wait(self, time_sec):
        '''
            Async wait for 'time_sec' seconds
        '''
        return self.env.timeout(time_sec)

    def async_wait_for_pkt(self, time_sec=None, limit=None):
        '''
            Wait for packet arrival for 'time_sec'

            if 'time_sec' is None will wait infinitly.
            if 'time_sec' is zero it will return immeaditly.

            if 'limit' is a number, it will return up to 'limit' packets
            even if there are more

            returns:
                list of packets
                each packet is a dict:
                    'pkt' - scapy packet
                    'ts'  - arrival TS(server time)

        '''
        return self.pkt.get(time_sec, limit)

    def async_tx_pkt(self, tx_pkt):
        '''
            Called by the sender side
            to transmit a packet

            'tx_pkt' - pkt as a binary to send

            call can choose to yield for TX actual
            event or ignore

            returns:
                dict:
                    'ts' - TX timestamp(server time)
        '''
        return self.tx_buffer.push(tx_pkt)

# ################## internal functions ##########################

    def _on_rx_pkt(self, pkt, rx_ts):
        '''
            Called by the reciver side
           (the service)
        '''
        self.pkt.put({'pkt': pkt, 'ts': rx_ts})
